#include "ivc_client.h"

ivcClient::ivcClient(domid_t domid,
                     uint16_t port,
                     const grant_ref_t *grefs,
                     uint32_t num_grants,
                     evtchn_port_t evtport,
                     eventController &e,
                     uint64_t connection_id)
    : mDomid(domid),
      mEvtchnPort(evtport),
      mEventController(e),
      mLog("ivcClient") {
    mLog.setStreamBuffer(&gSyslog);
    XenBackend::XenGnttabBuffer interimBuffer(domid, grefs, 32);

    mMappedBuffer = std::make_shared<XenBackend::XenGnttabBuffer>(domid, (grant_ref_t *)interimBuffer.get(), num_grants);

    mClient = (struct libivc_client *)malloc(sizeof(struct libivc_client));
    memset((void*)mClient, 0x00, sizeof(struct libivc_client));
    mClient->context = (void*)this;
    mClient->buffer = (char *)mMappedBuffer->get();
    mClient->num_pages = mMappedBuffer->size()/4096;
    mClient->remote_domid = domid;
    mClient->port = port;
    mClient->connection_id = connection_id;
    mEventCallback = std::function<void()>([&](){ eventCallback(); });
    mClient->event_channel = e.openEventChannel(domid, evtport, mEventCallback);
    mRingbuffer = std::make_shared<ringbuf>((uint8_t*)mClient->buffer, 4096 * num_grants, true);

    LOG(mLog, DEBUG) << "New client: " << "dom" << domid << ":" << port << "evtchn:" << evtport << "(" << mClient->event_channel << ")";
}

ivcClient::~ivcClient() {
    disconnectCallback();

    mRingbuffer = nullptr;
    if(mRingbuffer.use_count()) {
        LOG(mLog, ERROR) << "BUG: mRingbuffer has open references.\n";
    }

    mEventController.closeEventChannel(mClient->event_channel);
    mEventCallback = nullptr;

    mClient->port = 0;
    mClient->remote_domid = 0;
    mClient->buffer = nullptr;
    mClient->num_pages = 0;
    free(mClient);

    mMappedBuffer = nullptr;
    if(mMappedBuffer.use_count()) {
        LOG(mLog, ERROR) << "BUG: mMappedBuffer has open references.\n";
    }
}

void
ivcClient::disconnectCallback() {
    if(mClient && mClientDisconnectCallback) {
        mClientDisconnectCallback(mClient->opaque, mClient);
    }
}

void
ivcClient::eventCallback() {
    if(mClient && mClientEventCallback) {
        mClientEventCallback(mClient->opaque, mClient);
    }
}

#define RETRY_COUNT 5

int
ivcClient::recv(char *buf, uint32_t len) {
    unsigned int rc = 0;

    if (!mRingbuffer) {
        return -ENOENT;
    }

    rc = mRingbuffer->bytesAvailableRead();
    if (rc < len) {
        return -ENODATA;
    }

    rc = mRingbuffer->read((uint8_t*)buf, len);

    return len == rc ? 0 : -ENODATA;
}

int
ivcClient::send(char *buf, uint32_t len) {
    int rc;

    if (!mRingbuffer) {
        return -ENOENT;
    }

    rc = mRingbuffer->write((uint8_t*)buf, len);
    LOG(mLog, DEBUG) << "ivcClient::send: send notification to event-channel: " << mClient->event_channel << "(" << mEvtchnPort << ")";
    mEventController.notify(mClient->event_channel);

    return rc;
}

int
ivcClient::availableData() {
    return mRingbuffer->bytesAvailableRead();
}

int
ivcClient::availableSpace() {
    return mRingbuffer->bytesAvailableWrite();
}

void
ivcClient::setClientEventCallback(std::function<void(void *, libivc_client *)> fn) {
    mClientEventCallback = fn;
}

void
ivcClient::setClientDisconnectCallback(std::function<void(void *, libivc_client *)> fn) {
    mClientDisconnectCallback = fn;
}

void
ivcClient::setClientData(void *opaque) {
    mClient->opaque = opaque;
}

struct libivc_client *
ivcClient::client() {
    return mClient;
}

/*
 * Local variables:
 * mode: C++
 * c-file-style: "BSD"
 * c-basic-offset: 4
 * tab-width: 4
 * indent-tabs-mode: nil
 * End:
 */
